package o3;
import o3.parser.*;
import o3.lexer.*;
import o3.node.*;
import java.io.*;
import java.util.*;

public class Compiler {
  public static void main(String[] args) throws Exception {
    String fn = null, main = null;
    boolean dump = false;
    for(int i = 0; i < args.length; ++i) {
      if(args[i].equals("-d")) {
        dump = true;
      } else if(args[i].startsWith("-m")) {
        main = args[i].substring(2);
      } else if(args[i].startsWith("-")) {
        perror("uknown optios: " + args[i]);
      } else {
        fn = args[i]; break;
      }
    }
    if(fn == null) { perror("source filename not specified."); }
    Parser p = new Parser(new Lexer(new PushbackReader(
      new InputStreamReader(new FileInputStream(fn),"JISAutoDetect"),1024)));
    Start tree = p.parse();
    ClangEmit cle = new ClangEmit("code.c");
    TypeEnv tenv = new TypeEnv();
    Traverser trav = new Traverser(tenv, cle);
    tree.apply(trav);
    tenv.check(dump);
    tenv.cemit(cle);
    if(main != null) { emitmain(tenv, cle, main); }
    cle.close();
  }
  private static void emitmain(TypeEnv tenv, ClangEmit cle, String m) {
    String[] a = m.split("\\$");
    if(a.length != 3 || a[0].length() > 0 || a[2].length() == 0) {
      System.err.printf("-m%s should be -m$TYPE$meth.\n", m); return;
    }
    int tid = tenv.idn2tid(a[1]);
    Tbody.Base b = tenv.tid2tbody(tid);
    if(b == null || b.isGen() || !(b instanceof Tbody.Clus)) {
      System.err.printf("$%s$ is not a proper type.\n", a[1]); return;
    }
    Proctype p = ((Tbody.Clus)b).getProc(a[2]);
    if(p == null || p.getPara().length > 0) {
      System.err.printf("method %s undef or param# > 0.\n", a[2]); return;
    }
    cle.put("int main() {");
    cle.put(String.format(" %s_%s();", tenv.tid2sname(tid), a[2]));
    cle.put(" return 0;");
    cle.put("}");
  }
  private static void perror(String s) {
    System.err.println(s); System.exit(1);
  }
}
