package o3;
import java.util.*;
import java.lang.reflect.*;

public class Mtree {
  public interface Visitor {
    public void visit(Base b);
  }
  public static abstract class Base {
    int lineno = 0;
    public void show(int i) { System.out.println(nsp(i)+toString()); }
    public String toString() { return "?"; }
    public void exec(TpEquate te) { }
    public int getLineno() { return lineno; }
    public void perror(String s) { System.err.println(lineno+": "+s); }
  }
  public static class Ifst extends Base {
    Base[] list;
    public Ifst(int i, Base[] l) { lineno = i; list = l; }
    public void exec(TpEquate te) {
      for(Base b: list) {
        if(b instanceof Then) { if(((Then)b).execThen(te)) { return; } }
	else { ((Else)b).exec(te); }
      }
    }
    public void show(int i) {
      System.out.println(nsp(i) + "if:");
      for(Base b: list) { b.show(i+1); }
    }
  }
  public static class Then extends Base {
    Mexp cond;
    Base[] body;
    public Then(int l, Mexp c, Base[] b) {lineno=l;cond=c;body=b;}
    public boolean execThen(TpEquate te) {
      Mval v = cond.eval(te);
      if(!(v instanceof Vbool)) { perror("cond not bool."); return false; }
      if(!((Vbool)v).getValue()) { return false; }
      for(Base b: body) { b.exec(te); }
      return true;
    }
    public void show(int i) {
      System.out.println(nsp(i+1) + cond + " =>");
      for(Base b: body) { b.show(i+2); }
    }
  }
  public static class Else extends Base {
    Base[] body;
    public Else(int l, Base[] b) { lineno = l; body = b; }
    public void exec(TpEquate te) { for(Base b: body) { b.exec(te); } }
    public void show(int i) {
      System.out.println(nsp(i+1) + "true =>");
      for(Base b: body) { b.show(i+2); }
    }
  }
  public static class For extends Base {
    String idn = "?";
    Mexp expr = null;
    Base[] body;
    public For(int l, String i, Mexp e, Base[] b) {
      lineno = l; idn = i; expr = e; body = b;
    }
    public void exec(TpEquate te) {
      Mval v = expr.eval(te);
      if(v instanceof Vpset) {
        List<Vproc> list = ((Vpset)v).getValue();
        for(Vproc p: list) {
          TpEquate tp = new TpEquate(te); tp.add(idn, p);
          for(Base b: body) { b.exec(tp); }
        }
      } else if(v instanceof Vtset) {
        Set<Integer> set = ((Vtset)v).getValue();
        for(Integer t: set) {
          TpEquate tp = new TpEquate(te);
          tp.add(idn, new Vtype(te.getTe().tid2tbody(t)));
          for(Base b: body) { b.exec(tp); }
        }
      } else {
        perror(String.format("%s :not a set.", expr.toString()));
      }
    }
    public void show(int i) {
      System.out.printf("%sfor %s: %s\n", nsp(i), idn, expr.toString());
      for(Base b: body) { b.show(i+1); }
    }
  }
  public static class Msubr extends Base {
    static Map<String,Method> mtab = new HashMap<String,Method>();
    static Mval mnull = new Vnull();
    TpEquate te;
    static {
      Method[] a = Msubr.class.getDeclaredMethods();
      for(Method m: a) {
        String n = m.getName();
	if(n.startsWith("do_")) { mtab.put(n.substring(3), m); }
      }
    }
    public Msubr(int l, TpEquate te1) { lineno = l; te = te1; }
    public Method getMethod(String idn) { return mtab.get(idn); }
    public Mval do_add_super(Mval r, Mval x) {
      try{ 
        if(!(r instanceof Vtype) || !(x instanceof Vtype)) {
          perror("usage: type!add_super[type]:"+r+":"+x); return mnull;
        }
        int t1 = ((Vtype)r).getValue().getTid();
        int t2 = ((Vtype)x).getValue().getTid();
        te.getTe().addSup(t1, t2);
      } catch(Exception ex) { perror("sys:add_super:"+ex); }
      return mnull;
    }
    public Mval do_add_proctype(Mval r, Mval x) {
      try{ 
        if(!(r instanceof Vtype) || !(x instanceof Vproc)) {
          perror("usage: type!add_proctype[proc]:"+r+":"+x); return mnull;
        }
        Tbody.Base b = ((Vtype)r).getValue();
        Vproc p = (Vproc)x;
        Proctype q = p.getValue().copy(te.getTe());
        q.modRecv(p.getTid(), b.getTid());
        if(p.getIdn().equals("create")) { q.modRet(b.getTid()); }
        b.addProc(p.getIdn(), q);
      } catch(Exception ex) { perror("sys:add_proctype:"+ex); }
      return mnull;
    }
    public Mval do_add_proctypes(Mval r, Mval x) {
      try{ 
        if(!(r instanceof Vtype) || !(x instanceof Vtype)) {
          perror("usage: type!add_proctypes[type]:"+r+":"+x); return mnull;
        }
        Tbody.Base b = ((Vtype)r).getValue();
        Tbody.Base c = ((Vtype)x).getValue();
        for(String s: c.getPnames()) {
          Proctype p = c.getProc(s).copy(te.getTe());
          if(s.equals("create")) { p.modRet(b.getTid()); }
          else { p.modRecv(c.getTid(), b.getTid()); }
          b.addProc(s, p);
        }
      } catch(Exception ex) { perror("sys:add_proctypes:"+ex); }
      return mnull;
    }
    public Mval do_add_procdef(Mval r, Mval x) {
      try{ 
        if(!(r instanceof Vtype) || !(x instanceof Vproc)) {
          perror("usage: type!add_procdef[proc]:"+r+":"+x); return mnull;
        }
        Tbody.Base b = ((Vtype)r).getValue();
        Vproc p = (Vproc)x;
        Proctype q = b.getProc(p.getIdn());
        q.putBody(p.getValue().getBody());
      } catch(Exception ex) { perror("sys:add_procdef"+ex); }
      return mnull;
    }
    public Mval do_add_procdefs(Mval r, Mval x) {
      try{ 
        if(!(r instanceof Vtype) || !(x instanceof Vtype)) {
          perror("usage: type!add_procdef[proc]:"+r+":"+x); return mnull;
        }
        Tbody.Base b = ((Vtype)r).getValue();
        Tbody.Base c = ((Vtype)x).getValue();
        for(String s: c.getPnames()) {
          Proctype p = b.getProc(s);
          Proctype q = c.getProc(s);
          if(p != null && q != null) { p.putBody(q.getBody()); }
        }
      } catch(Exception ex) { perror("sys:add_procdefs:"+ex); }
      return mnull;
    }
    public Mval do_procdefs(Mval r) { 
      try {
        if(!(r instanceof Vtype)) {
          perror("usage: type!procdefs[]:"+r); return mnull;
        }
        Tbody.Base b = ((Vtype)r).getValue();
        Vpset pset = new Vpset();
        for(String s: b.getPnames()) {
          pset.add(new Vproc(b.getTid(), s, b.getProc(s)));
        }
        return pset;
      } catch(Exception ex) { perror("sys:procdefs:"+ex); }
      return mnull;
    }
    public Mval do_name_matches(Mval p, Mval q) {
      try {
        if(!(p instanceof Vproc) || !(q instanceof Vstr)) {
          perror("usage: proc!name_matches[str]:"+p+":"+q); return mnull;
        }
        return new Vbool(((Vproc)p).getIdn().matches(((Vstr)q).getValue()));
      } catch(Exception ex) { perror("sys:name_matches:"+ex); }
      return mnull;
    }
    public Mval do_add_body_before(Mval p, Mval q) {
      try {
        if(!(p instanceof Vproc) || !(q instanceof Vproc)) {
          perror("usage: proc!add_body_after[proc]:"+p+":"+q); return mnull;
        }
        Proctype pp = ((Vproc)p).getValue();
        Proctype qp = ((Vproc)q).getValue();
        if(((Vproc)p).getIdn().equals("create")) {
          perror("cannot add before daemon to create"); return mnull;
        }
        pp.putBodyBefore(qp);
      } catch(Exception ex) { perror("sys:add_body_before:"+ex); }
      return mnull;
    }
    public Mval do_add_body_after(Mval p, Mval q) {
      try {
        if(!(p instanceof Vproc) || !(q instanceof Vproc)) {
          perror("usage: proc!add_body_after[proc]:"+p+":"+q); return mnull;
        }
        int t = ((Vproc)p).getTid();
        Proctype pp = ((Vproc)p).getValue();
        Proctype qp = ((Vproc)q).getValue().copyWithBody(te.getTe());
        String msg = null;
        if(((Vproc)p).getIdn().equals("create")) {
          msg = qp.chkAddCreate(te.getTe(), t, pp);
          qp.getBody().getVarEnv().moveToParam("self", 0);
        } else {
          msg = qp.chkAdd(te.getTe(), t, pp);
        }
        if(msg != null) {
          perror(msg + ":" + qp.toString() + ":" + pp.toString());
        } else {
          pp.putBodyAfter(qp);
        }
      } catch(Exception ex) { perror("sys:add_body_after:"+ex); }
      return mnull;
    }
  }
  public static abstract class Mexp extends Base {
    static Mval mnull = new Vnull();
    public void exec(TpEquate te1) { eval(te1); }
    public abstract Mval eval(TpEquate te);
  }
  public static class Mcall extends Mexp {
    String meth = "?";
    Mexp[] args;
    public Mcall(int l, Mexp r, String m, Mexp[] a) {
      lineno = l; meth = m; args = new Mexp[a.length+1]; args[0] = r;
      for(int i = 0; i < a.length; ++i) { args[i+1] = a[i]; }
    }
    public Mval eval(TpEquate te) {
      int mtid; Method m; boolean ok = true;
      Msubr ms = new Msubr(lineno, te);
      int[] a = new int[args.length];
      Mval[] b = new Mval[args.length];
      for(int i = 0; i < args.length; ++i) { b[i] = args[i].eval(te); }
      if((mtid = te.getTe().lookMproc(meth)) >= 0) {
        Tbody.Mclu mproc = (Tbody.Mclu)te.getTe().tid2tbody(mtid);
        for(int i = 0; i < args.length; ++i) {
          if(!(b[i] instanceof Vtype)) {
            perror(String.format("arg %d of %s not a type",
                                 i, this.toString())); ok = false;
          } else { a[i] = ((Vtype)b[i]).getTid(); }
        }
        if(ok && mproc.getNpara() != a.length) {
          perror("#args mismatch for: "+this.toString()); ok = false;
        }
        if(ok) { mproc.callProc(te.getTe().getTenv(),mproc.getParas(),a); }
        return mnull;
      } else if((m = ms.getMethod(meth)) == null) {
        perror("method "+meth+" not found"); return mnull;
      } else if(m.getParameterCount() != args.length) { 
        perror(String.format("%s expect %d args, given %d.",
	                      meth, m.getParameterCount(), args.length));
        return mnull;
      } else {
        try {
          return (Mval)m.invoke(ms, (Object[])b);
	} catch(Exception ex) {
	  perror("method call failed: "+m+":"+ex.toString()); return mnull; 
	}
      }
    }
    public String toString() {
      String s = String.format("%s!%s[", args[0].toString(), meth);
      for(int i = 1; i < args.length; ++i) {
        if(i > 1) { s += ","; }
	s += args[i].toString();
      }
      return s + "]";
    }
  }
  public static class Mtype extends Mexp {
    TypeName tname;
    public Mtype(int l, TypeName t) { lineno = l; tname = t; }
    public TypeName getTypename() { return tname; }
    public Mval eval(TpEquate te) {
      if(tname.getNpara() == 0) { return te.get(tname.getIdn()); }
      int tid = te.getTe().tname2tid(tname);
      if(tid < 0) { perror("undef type: "+tname); return mnull; }
      return new Vtype(te.getTe().tid2tbody(tid));
    }
    public String toString() { return tname.toString(); }
  }
  public static class Mproc extends Mexp {
    TypeName tname;
    String pname;
    public Mproc(int l, TypeName t, String p) {
      lineno=l; tname = t; pname = p;
    }
    public TypeName getTypename() { return tname; }
    public Mval eval(TpEquate te) {
      int tid = te.getTe().tname2tid(tname);
      if(tid < 0) { perror("undef type: "+tname); return mnull; }
      Tbody.Base b = te.getTe().tid2tbody(tid);
      Proctype p = b.getProc(pname);
      if(p == null) { perror("proc "+pname+" not in "+tname); return mnull; }
      return new Vproc(tid, pname, p);
    }
    public String getPname() { return pname; }
    public String toString() { return "$"+tname+"$"+pname; }
  }
  public static class Mstr extends Mexp {
    String value = "?";
    public Mstr(int l, String s) { lineno = l; value = s; }
    public Mval eval(TpEquate te) { return new Vstr(value); }
    public String toString() { return value; }
  }
  public static abstract class Mval {
  }
  public static class Vnull extends Mval {
    public String toSring() { return null; }
  }
  public static class Vbool extends Mval {
    boolean value;
    public Vbool(boolean b) { value = b; }
    public boolean getValue() { return value; }
    public String toString() { return ""+value; }
  }
  public static class Vstr extends Mval {
    String value;
    public Vstr(String s) { value = unesc(s); }
    public String getValue() { return value; }
    public String toString() { return '"' + value + '"'; }
  }
  public static class Vtype extends Mval {
    Tbody.Base body;
    public Vtype(Tbody.Base b) { body = b; }
    public Tbody.Base getValue() { return body; }
    public int getTid() { return body.getTid(); }
    public String toString() { return "tid:"+body.getTid(); }
  }
  public static class Vproc extends Mval {
    int tid;
    String idn;
    Proctype value;
    public Vproc(int t, String i, Proctype p) { tid = t; idn = i; value = p; }
    public int getTid() { return tid; }
    public String getIdn() { return idn; }
    public Proctype getValue() { return value; }
    public String toString() { return value.toString(); }
  }
  public static class Vtset extends Mval {
    Set<Integer> set = new TreeSet<Integer>();
    public void add(Integer i) { set.add(i); }
    public Set<Integer> getValue() { return set; }
  }
  public static class Vpset extends Mval {
    List<Vproc> list = new ArrayList<Vproc>();
    public void add(Vproc p) { list.add(p); }
    public List<Vproc> getValue() { return list; }
  }
  private static String nsp(int i) {
    String s = ""; while(s.length() < i) { s+=" "; } return s;
  }
  private static String unesc(String s) {
    s = s.substring(1, s.length()-1);
    int len = s.length();
    StringBuilder b = new StringBuilder();
    for(int i = 0; i < len; ++i) {
      if(s.charAt(i) != '\\') {
        b.append(s.charAt(i));
      } else if(i+1 < len) {
        b.append(s.charAt(++i));
      }
    }
    return b.toString();
  }
  public static class TpEquate {
    static Mval mnull = new Vnull();
    TypeEquate te = null;
    TpEquate tp = null;
    Map<String,Mval> map = new TreeMap<String,Mval>();
    public TpEquate(TypeEquate te1) { te = te1; }
    public TpEquate(TpEquate tp1) { tp = tp1; }
    public TypeEquate getTe() { return (te != null) ? te : tp.getTe(); }
    public void add(String s, Mval v) { map.put(s, v); }
    public Mval get(String s) {
      if(map.containsKey(s)) { return map.get(s); }
      if(tp != null) { return tp.get(s); }
      int t = te.tname2tid(new TypeName(s));
      return (t < 0) ? mnull : new Vtype(te.tid2tbody(t));
    }
  }
}
