package o3;
import java.util.*;

public class Proctype {
  TypeEquate te;
  int[] para;
  int rtype;
  boolean inst = false;
  Tree.Proc body;
  Tree.Vdef[] vdefs = null;
  List<Proctype> aft = new ArrayList<Proctype>();
  List<Proctype> bef = new ArrayList<Proctype>();
  String str = "?";
  public Proctype(TypeEquate e, int[] p, int r, Tree.Proc b) {
    te = e; para = p; rtype = r; body = b; genstr();
  }
  private void genstr() {
    str = inst ? "proc*(" : "proc(";
    if(para.length > 0) { str += te.tid2sname(para[0]); }
    for(int i = 1; i < para.length; ++i) {
      str += "," + te.tid2sname(para[i]);
    }
    str += (rtype<0) ? ")" : ("):" + te.tid2sname(rtype));
  }
  public boolean isInst() { return inst; }
  public void putInst(boolean b) { inst = b; }
  public int[] getPara() { return para; }
  public int getRtype() { return rtype; }
  public Tree.Proc getBody() { return body; }
  public void putVdefs(Tree.Vdef[] v) { vdefs = v; }
  public void collectVar(String s, VarEnv ve) {
    colv(s, ve, this);
    for(Proctype p: bef) { colv(s, ve, p); }
    for(Proctype p: aft) { colv(s, ve, p); }
  }
  private void colv(String e, VarEnv ve1, Proctype q) {
    VarEnv ve2 = q.getBody().getVarEnv();
    String e2 = q.getBody().toString();
    for(String s: ve2) {
      int p = ve2.getPosition(s); if(p != -1) { continue; }
      if(!ve1.hasName(s)) { ve1.add(s, ve2.getTypename(s), -1); continue; }
      if(ve1.getPosition(s) != -1) {
        System.err.printf("%s <- %s: conf: param vs instv: %s\n", e, e2, s);
        continue;
      }
      int t1 = te.tname2tid(ve1.getTypename(s));
      int t2 = te.tname2tid(ve2.getTypename(s));
      if(t1 != t2) {
        System.err.printf("%s <- %s: conf: %s's type %s vs %s\n", e, e2,
                          te.tid2cname(t1), te.tid2cname(t2));
      }
    }
  }
  public void putBody(Tree.Proc b) { body = b; }
  public void putBodyAfter(Proctype b) { aft.add(b); }
  public void putBodyBefore(Proctype b) { bef.add(b); }
  public List<Proctype> getBodyBefore() { return bef; }
  public List<Proctype> getBodyAfter() { return aft; }
  public void modRet(int tid) { rtype = tid; genstr(); }
  public void modRecv(int t1, int t2) {
    if(para.length > 0 && para[0] == t1) { para[0]=t2; inst=true; genstr(); }
  }
  public Proctype copy(TypeEquate e) {
    int[] pa = new int[para.length];
    for(int i = 0; i < para.length; ++i) { pa[i] = para[i]; }
    Proctype p = new Proctype(e, pa, rtype, null);
    p.putVdefs(vdefs); p.putInst(inst); return p;
  }
  public Proctype copyWithBody(TypeEquate e) {
    int[] pa = new int[para.length];
    for(int i = 0; i < para.length; ++i) { pa[i] = para[i]; }
    Proctype p = new Proctype(e, pa, rtype, body.copy());
    p.putVdefs(vdefs); p.putInst(inst); return p;
  }
  public String chkAdd(TypeEquate te1, int tid, Proctype q) {
    TypeEnv te = te1.getTenv();
    int[] para1 = q.getPara();
    if(para.length > para1.length) { return "ext. meth with excess args"; }
    for(int i = 0; i < Math.min(para.length, para1.length); ++i) {
      if(i == 0 && q.isInst()) {
        if(!inst) { return "inst. meth extended by non-inst one"; }
        para[0] = tid;
      } else {
        if(!te.isSup(para[i], para1[i])) { return "incompat ext. arg "+i; }
      }
    }
    return null;
  }
  public String chkAddCreate(TypeEquate te1, int tid, Proctype q) {
    TypeEnv te = te1.getTenv();
    int[] para1 = q.getPara();
    if(para.length > para1.length) { return "ext. meth with excess args"; }
    for(int i = 0; i < Math.min(para.length, para1.length); ++i) {
      if(!te.isSup(para[i], para1[i])) { return "incompat ext. arg "+i; }
    }
    int[] para2 = new int[para.length+1];
    para2[0] = tid;
    for(int i = 0; i < para.length; ++i) { para2[i+1] = para[i]; }
    para = para2; return null;
  }
  public String toString() { return str; }
  public void cemit_dcl(ClangEmit cle, int tid, String pfx, String idn) {
    cle.put(fhdr(cle, tid, pfx, idn)+";");
  }
  public void show(int i) {
    System.out.println(nsp(i)+str);
    for(int k = bef.size()-1; k >= 0; --k) {
      System.out.printf("%s<Before %d>\n", nsp(i+1), k);
      bef.get(k).show(i+2);
    }
    body.show(i+1);
    for(int k = 0; k < aft.size(); ++k) {
      System.out.printf("%s<After %d>\n", nsp(i+1), k);
      aft.get(k).show(i+2);
    }
  }
  public void cemit_body(ClangEmit cle, int tid, String pfx, String idn,
                         VarEnv ve) {
    boolean genself = false;
    String fn = pfx+"_"+idn;
    cle.put(fhdr(cle, tid, pfx, idn)+" {");
    if(idn.equals("create")) {
      genself = true;
//    cle.put(String.format("extern vtbl_r %s_vtbl[];", pfx));
      cle.put(String.format("%s_p self = (%s_p)malloc(sizeof(%s_r));",
                            pfx, pfx, pfx));
      cle.put(String.format("self->ops = %s_vtbl;", pfx));
    }
    if(vdefs != null) { for(Tree.Vdef v: vdefs) { v.cemit(cle, ve); } }
    body.cemit2(te, cle, tid, pfx, genself);
    cle.put("}");
  }
  private String fhdr(ClangEmit cle, int tid, String pfx, String idn) {
    String fn = pfx+"_"+idn, ret = (rtype>=0) ? "obj_p":"void";
    ret += String.format(" %s(", fn);
    int d = para.length > body.getParas().length ? -1 : 0;
    for(int i = 0; i < para.length; ++i) {
      if(i == 0 && d < 0) {
        ret += te.tid2sname(para[i]) + "_p self";
      } else if(i == 0 && inst) {
        ret += te.tid2sname(para[i]) + "_p " + body.getParas()[i].getName();
      } else if(i == 0) {
        ret += "obj_p " + body.getParas()[i].getName();
      } else {
        ret += ", obj_p " + body.getParas()[i+d].getName();
      }
    }
    return ret + ")";
  }
  private static String nsp(int i) {
    String s = ""; while(s.length() < i) { s+=" "; } return s;
  }
}
