package o3;

public class Tree {
  public interface Visitor {
    public void visit(Base b);
  }
  public static class Typecheck implements Visitor {
    TypeEquate te;
    VarEnv ve;
    public Typecheck(TypeEquate te1, VarEnv ve1) { te = te1; ve = ve1; }
    public void visit(Base b) {
      if(b instanceof Expr) { ((Expr)b).tcheck(te, ve); }
      if(b instanceof Vdef) { ((Vdef)b).tcheck(te, ve); }
    }
  }
  public static abstract class Base {
    int lineno = 0;
    public void show(int i) { System.out.println(nsp(i)+toString()); }
    public String toString() { return "?"; }
    public void accept(Visitor v) { v.visit(this); }
    public int getLineno() { return lineno; }
    public void cemit(ClangEmit cle, VarEnv ve) { }
  }
  public static class Vdef extends Base {
    String idn;
    TypeName typ;
    Expr expr = null;
    public Vdef(int l, String i, TypeName t) {lineno=l;idn=i;typ=t; }
    public Vdef(int l, String i, TypeName t, Expr e) {
      this(l, i, t); expr = e;
    }
    public String getName() { return idn; }
    public TypeName getTypename() { return typ; }
    public Expr getExpr() { return (Expr)expr; }
    public String toString() {
      String s = idn + ":" + typ;
      if(expr != null) { s += " := " + expr.toString(); }
      return s;
    }
    public void accept(Visitor v) {
      if(expr != null) { expr.accept(v); } v.visit(this);
    }
    public void tcheck(TypeEquate te, VarEnv ve) {
      if(expr == null) { return; }
      int tid = te.tname2tid(ve.getTypename(idn));
      if(tid != expr.getTypeid()) { 
        expr.perror(String.format("assign compat: LHS=%s RHS=%s",
          te.tid2cname(tid), te.tid2cname(expr.getTypeid())));
      }
    }
    public void cemit(ClangEmit cle, VarEnv ve) {
      if(expr == null) { return; }
      int p = ve.getPosition(idn);
      String lhs = (p < -1 || p >= 0) ? idn : "self->"+idn;
      cle.put(lhs+"="+expr.cemits(ve)+";");
    }
  }
  public static class Proc extends Base {
    String idn;
    TypeName rtype;
    Annots annot;
    Base[] body;
    Para[] para;
    VarEnv ve = new VarEnv();
    public Proc(int l, String i, Annots a, Base[] b, Para[] p) {
      lineno = l; idn = i.intern(); annot = a; body = b; para = p;
    }
    public Proc(int l, String i, Annots a, Base[] b, Para[] p, TypeName r) {
      this(l, i, a, b, p); rtype = r;
    }
    public Proc copy() {
      Proc p = new Proc(lineno, idn, annot, body, para, rtype);
      p.setVarEnv(ve.copy()); return p;
    }
    public void check1(TypeEquate te, VarEnv ve1, TypeName tn) {
      int i = 0;
      for(Para p: para) { ve.add(p.getName(), p.getTypename(), i++); }
      if(idn.equals("create")) { ve.add("self", tn, -2); }
      for(Base b: body) {
        b.accept(new Visitor() {
          public void visit(Base b) {
            if(!(b instanceof Vdef)) { return; }
            Vdef v = (Vdef)b; ve.add(v.getName(), v.getTypename(), -2);
          }
        });
      }
      ve1.copyTo(ve);
    }
    public String getIdn() { return idn; }
    public Para[] getParas() { return para; }
    public TypeName getRtype() { return rtype; }
    public VarEnv getVarEnv() { return ve; }
    public void setVarEnv(VarEnv ve1) { ve = ve1; }
    public void accept(Visitor v) {
      v.visit(this);
      for(Base b: body) { b.accept(v); }
    }
    public String toString() {
      String s = idn + " = proc(";
      if(para.length > 0) {
        s += para[0].toString();
        for(int i = 1; i < para.length; ++i) { s +=","+para[i].toString(); }
      }
      s += ")";
      if(rtype != null) { s += ":" + rtype; }
      return s;
    }
    public void show(int i) {
      System.out.println(nsp(i) + toString());
      System.out.println(nsp(i+1) + ve.toString());
      annot.show(i+2);
      for(Base b: body) { b.show(i+2); }
    }
    public void cemit2(TypeEquate te, ClangEmit cle, int tid,
                       String pfx, boolean skipself) {
      StrList list = annot.get("@clangemit");
      if(list != null) {
        for(String s:list) {
          if(s.charAt(0) == '"') {
            String t = unesc(s);
            if(t.indexOf("%s") >= 0) {
              cle.put(String.format(t, pfx, pfx, pfx, pfx));
            } else {
              cle.put(String.format(t, tid, tid, tid, tid));
            }
          }
        }
      } else {
        for(String v: ve) {
          if(ve.getPosition(v) >= -1) { continue; }
          if(skipself && v.equals("self")) { continue; }
          int t = te.tname2tid(ve.getTypename(v));
          if(t < 0) { t = TypeEnv.NONE; }
//        String s = te.tid2sname(t);
          cle.put(String.format("obj_p %s;", v));
        }
        for(Base b: body) { b.cemit(cle, ve); }
      }
    }
  }
  public static class Para extends Base {
    String str;
    TypeName typ;
    public Para(int l, String s, TypeName t) { lineno = l; str = s; typ = t; }
    public String getName() { return str; }
    public TypeName getTypename() { return typ; }
    public String toString() { return str + ":" + typ; }
  }
  public static class Retn extends Base {
    Expr expr;
    public Retn(int l, Expr e) { lineno = l; expr = e; }
    public void accept(Visitor v) { v.visit(this); expr.accept(v); }
    public String toString() { return (expr==null)?"return":"return "+expr; }
    public void cemit(ClangEmit cle, VarEnv ve) {
      if(expr == null) { cle.put("return;"); }
      else             { cle.put("return (obj_p)"+expr.cemits(ve)+";"); }
    }
  }
  public static class Whst extends Base {
    Expr cond;
    Base[] body;
    public Whst(int l, Expr c, Base[] b) { lineno = l; cond = c; body = b; }
    public void accept(Visitor v) {
      cond.accept(v); v.visit(this);
      for(Base b: body) { b.accept(v); }
    }
    public void show(int i) {
      System.out.println(nsp(i) + "while: " + cond);
      for(Base t: body) { t.show(i+1); }
    }
    public void cemit(ClangEmit cle, VarEnv ve) {
      cle.put("while("+cond.cemits(ve)+"==(obj_p)(&bool_true)) {");
      for(Base b: body) { b.cemit(cle, ve); }
      cle.put("}");
    }
  }
  public static class Ifst extends Base {
    Base[] list;
    public Ifst(int i, Base[] l) { lineno = i; list = l; }
    public void accept(Visitor v) {
      v.visit(this);
      for(Base b: list) { b.accept(v); }
    }
    public void show(int i) {
      System.out.println(nsp(i) + "if:");
      for(Base b: list) { b.show(i+1); }
    }
    public void cemit(ClangEmit cle, VarEnv ve) {
      cle.put("if(0) {");
      cle.put("}");
      for(Base b: list) { b.cemit(cle, ve); }
    }
  }
  public static class Then extends Base {
    Expr cond;
    Base[] body;
    public Then(int l, Expr c, Base[] b) { lineno = l; cond = c; body = b; }
    public void accept(Visitor v) {
      cond.accept(v); v.visit(this);
      for(Base b: body) { b.accept(v); }
    }
    public void show(int i) {
      System.out.println(nsp(i+1) + cond + " =>");
      for(Base b: body) { b.show(i+2); }
    }
    public void cemit(ClangEmit cle, VarEnv ve) {
      cle.put("else if("+cond.cemits(ve)+"==(obj_p)(&bool_true)) {");
      for(Base b: body) { b.cemit(cle, ve); }
      cle.put("}");
    }
  }
  public static class Else extends Base {
    Base[] body;
    public Else(int l, Base[] b) { lineno = l; body = b; }
    public void accept(Visitor v) {
      v.visit(this);
      for(Base b: body) { b.accept(v); }
    }
    public void show(int i) {
      System.out.println(nsp(i+1) + "true =>");
      for(Base b: body) { b.show(i+2); }
    }
    public void cemit(ClangEmit cle, VarEnv ve) {
      cle.put("else {");
      for(Base b: body) { b.cemit(cle, ve); }
      cle.put("}");
    }
  }
  public static abstract class Expr extends Base {
    int typeid = 0;
    public int getTypeid() { return typeid; }
    public abstract void tcheck(TypeEquate te, VarEnv ve);
    public abstract String cemits(VarEnv ve);
    public void cemit(ClangEmit cle, VarEnv ve){cle.put(this.cemits(ve)+";");}
    public void perror(String s) {
      System.err.printf("%d: "+s+" %s\n", lineno, this.toString());
    }
  }
  public static class Assign extends Expr {
    String idn;
    Expr rhs;
    public Assign(int l, String i, Base r) { lineno=l; idn=i; rhs=(Expr)r; }
    public String getIdn() { return idn; }
    public Base getRhs() { return rhs; }
    public void accept(Visitor v) { rhs.accept(v); v.visit(this); }
    public String toString() { return idn + " := " + rhs; }
    public void tcheck(TypeEquate te, VarEnv ve) {
      int tid = te.tname2tid(ve.getTypename(idn));
      if(tid != rhs.getTypeid()) {
        perror(String.format("assign compat: LHS=%s RHS=%s",
          te.tid2cname(tid), te.tid2cname(rhs.getTypeid())));
      }
      typeid = rhs.getTypeid();
    }
    public String cemits(VarEnv ve) {
      int p = ve.getPosition(idn);
      String lhs = (p < -1 || p >= 0) ? idn : "self->"+idn;
      return lhs+"="+rhs.cemits(ve);
    }
  }
  public static class Call extends Expr {
    Expr recv;
    String mesg = "?";
    Expr[] args = {};
    public Call(int l, Base r, String m) {
      lineno = l; recv = (Expr)r; mesg = m;
    }
    public Call(int l, Base r, String m, Base a1) {
      this(l, r, m); args = new Expr[]{(Expr)a1};
    }
    public Call(int l, Base r, String m, Base a1, Base a2) { 
      this(l, r, m); args = new Expr[]{(Expr)a1, (Expr)a2};
    }
    public Call(int l, Base r, String m, Expr[] a) { this(l,r,m); args = a; }
    public void accept(Visitor v) {
      recv.accept(v);
      for(Expr b: args) { b.accept(v); }
      v.visit(this);
    }
    public int getLineno() { return lineno; }
    public String toString() { 
      String s = recv + "!" + mesg + "(";
      if(args.length > 0) { s += args[0]; }
      for(int i = 1; i < args.length; ++i) { s += ", "+args[i]; }
      return s+")";
    }
    public void tcheck(TypeEquate te, VarEnv ve) {
      if(recv.getTypeid() < 0) { perror("unknown recv type"); return; }
      Proctype p = te.tid2tbody(recv.getTypeid()).getProc(mesg);
      if(p == null) { perror("unknown method"); return; }
      typeid = p.getRtype();
      int[] a = p.getPara();
      if(args.length != a.length-1) { perror("arg count mismatch"); }
      for(int i = 0; i < Math.min(args.length, a.length-1); ++i) {
        if(args[i].getTypeid() != a[i+1]) {
          perror(String.format("arg %d expect %s got %s", i+1,
                 te.tid2cname(a[i+1]), te.tid2cname(args[i].getTypeid())));
        }
      }
    }
    public String cemits(VarEnv ve) {
      String s="send"+args.length+"((obj_p)"+recv.cemits(ve)+",\""+mesg+"\"";
      for(int i = 0; i < args.length; ++i) { s += ","+args[i].cemits(ve); }
      return s+")";
    }
  }
  public static class Tcall extends Expr {
    TypeName typ;
    int pnum = 0;
    String mesg = "?", sname = "?";
    Expr[] args = {};
    public Tcall(int l, TypeName t, String m) {
      lineno = l; typ = t; mesg = m;
    }
    public Tcall(int l, TypeName t, String m, Base a1) {
      this(l, t, m); args = new Expr[]{(Expr)a1};
    }
    public Tcall(int l, TypeName t, String m, Base a1, Base a2) { 
      this(l, t, m); args = new Expr[]{(Expr)a1, (Expr)a2};
    }
    public Tcall(int l, TypeName t, String m, Expr[] a) {this(l,t,m);args=a;}
    public void accept(Visitor v) {
      for(Expr b: args) { b.accept(v); }
      v.visit(this);
    }
    public int getLineno() { return lineno; }
    public String toString() { 
      String s = "$" + typ + "$" + mesg + "(";
      if(args.length > 0) { s += args[0]; }
      for(int i = 1; i < args.length; ++i) { s += ", "+args[i]; }
      return s+")";
    }
    public void tcheck(TypeEquate te, VarEnv ve) {
      int tid = te.tname2tid(typ);
      Proctype p = te.tid2tbody(tid).getProc(mesg);
      if(p == null) { perror("undef method: "+mesg); return; }
      pnum = te.tid2tbody(tid).getPnum(mesg);
      typeid = p.getRtype();
      int[] a = p.getPara();
      sname = typ.getIdn()+"_"+tid;
      if(args.length != a.length) { perror("arg count mismatch"); }
      for(int i = 0; i < Math.min(args.length, a.length); ++i) {
        if(args[i].getTypeid() != a[i]) {
          perror(String.format("arg %d expect %s got %s", i+1,
                 te.tid2cname(a[i]), te.tid2cname(args[i].getTypeid())));
        }
      }
    }
    public String cemits(VarEnv ve) {
      String s = mesg.equals("create") ? "callx" : "call";
      s += args.length + "(" + sname+"_vtbl+"+pnum;
      for(int i = 0; i < args.length; ++i) {s+=", "+args[i].cemits(ve);}
      return s+")";
    }
  }
  public static class Not extends Expr {
    Expr expr;
    public Not(int l, Base t) { lineno = l; expr = (Expr)t; }
    public void accept(Visitor v) { expr.accept(v); v.visit(this); }
    public Expr getExpr() { return expr; }
    public String toString() { return "!(" + expr + ")"; }
    public void tcheck(TypeEquate te, VarEnv ve) {
      if(expr.getTypeid() != te.BOOL) { perror("OPD not bool."); }
      typeid = te.BOOL;
    }
    public String cemits(VarEnv ve) {
      return "(obj_p)("+expr.cemits(ve) +
             "==(obj_p)&bool_true?&bool_false:&bool_true)";
    }
  }
  public static class Cand extends Expr {
    Expr left, right;
    public Cand(int i, Base l, Base r) {
      lineno = i; left = (Expr)l; right = (Expr)r;
    }
    public void accept(Visitor v) {
      left.accept(v); right.accept(v); v.visit(this);
    }
    public Expr getLeft() { return left; }
    public Expr getRight() { return right; }
    public String toString() { return "(" + left + "&" + right + ")"; }
    public void tcheck(TypeEquate te, VarEnv ve) {
      if(left.getTypeid() != te.BOOL) { perror("LHS not bool."); }
      if(right.getTypeid() != te.BOOL) { perror("RHS not bool."); }
      typeid = te.BOOL;
    }
    public String cemits(VarEnv ve) {
      return "("+left.cemits(ve)+")==&bool_true&&("+right.cemits(ve)+"==&bool_true)";
    }
  }
  public static class Cor extends Expr {
    Expr left, right;
    public Cor(int i, Base l, Base r) {
      lineno = i; left = (Expr)l; right = (Expr)r;
    }
    public void accept(Visitor v) {
      left.accept(v); right.accept(v); v.visit(this); 
    }
    public Expr getLeft() { return left; }
    public Expr getRight() { return right; }
    public String toString() { return "(" + left + "||" + right + ")"; }
    public void tcheck(TypeEquate te, VarEnv ve) {
      if(left.getTypeid() != te.BOOL) { perror("LHS not bool."); }
      if(right.getTypeid() != te.BOOL) { perror("RHS not bool."); }
      typeid = te.BOOL;
    }
    public String cemits(VarEnv ve) {
      return "("+left.cemits(ve)+")==&bool_true||("+right.cemits(ve)+"==&bool_true)";
    }
  }
  public static class Ident extends Expr {
    String idn = "?";
    public Ident(int l, String s) { lineno = l; idn = s; }
    public String toString() { return idn; }
    public void tcheck(TypeEquate te, VarEnv ve) {
      typeid = te.tname2tid(ve.getTypename(idn));
    }
    public String cemits(VarEnv ve) {
      int p = ve.getPosition(idn);
      String lhs = (p < -1 || p >= 0) ? idn : "self->"+idn;
      return lhs;
    }
  }
  public static class Iconst extends Expr {
    int val = 0;
    public Iconst(int l, Integer i) {
      lineno = l; typeid = TypeEquate.INT; val = i;
    }
    public String toString() { return val+""; }
    public void tcheck(TypeEquate te, VarEnv ve) { typeid = te.INT; }
    public String cemits(VarEnv ve) {
      return String.format("int_3_create(%d)", val);
    }
  }
  public static class Sconst extends Expr {
    String str = "";
    public Sconst(int l, String s) {
      lineno = l; typeid = TypeEquate.STRING; str = s;
    }
    public String toString() { return str; }
    public void tcheck(TypeEquate te, VarEnv ve) { typeid = te.STRING; }
    public String cemits(VarEnv ve) { 
      return String.format("string_4_create(\"%s\")", unesc(str));
    }
  }
  public static class Bconst extends Expr {
    boolean val = false;
    public Bconst(int l, Boolean b) {
      lineno = l; typeid = TypeEquate.BOOL; val = b;
    }
    public String toString() { return val+""; }
    public void tcheck(TypeEquate te, VarEnv ve) { typeid = te.BOOL; }
    public String cemits(VarEnv ve) { return String.format("(&bool_%b)", val); }
  }
  public static class Nil extends Expr {
    public Nil(int l) { lineno = l; typeid = TypeEquate.NONE; }
    public String toString() { return "nil"; }
    public void tcheck(TypeEquate te, VarEnv ve) { typeid = te.NONE; }
    public String cemits(VarEnv ve) { return "0"; }
  }
  private static String nsp(int i) {
    String s = ""; while(s.length() < i) { s+=" "; } return s;
  }
  private static String unesc(String s) {
    s = s.substring(1, s.length()-1);
    int len = s.length();
    StringBuilder b = new StringBuilder();
    for(int i = 0; i < len; ++i) {
      if(s.charAt(i) != '\\') {
        b.append(s.charAt(i));
      } else if(i+1 < len) {
        b.append(s.charAt(++i));
      }
    }
    return b.toString();
  }
}
